# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import transaction_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2


class TransactionServiceStub(object):
    """Allows clients to read transactions from the ledger.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetTransactions = channel.unary_stream(
                '/com.daml.ledger.api.v1.TransactionService/GetTransactions',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsResponse.FromString,
                )
        self.GetTransactionTrees = channel.unary_stream(
                '/com.daml.ledger.api.v1.TransactionService/GetTransactionTrees',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionTreesResponse.FromString,
                )
        self.GetTransactionByEventId = channel.unary_unary(
                '/com.daml.ledger.api.v1.TransactionService/GetTransactionByEventId',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.FromString,
                )
        self.GetTransactionById = channel.unary_unary(
                '/com.daml.ledger.api.v1.TransactionService/GetTransactionById',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.FromString,
                )
        self.GetFlatTransactionByEventId = channel.unary_unary(
                '/com.daml.ledger.api.v1.TransactionService/GetFlatTransactionByEventId',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.FromString,
                )
        self.GetFlatTransactionById = channel.unary_unary(
                '/com.daml.ledger.api.v1.TransactionService/GetFlatTransactionById',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.FromString,
                )
        self.GetLedgerEnd = channel.unary_unary(
                '/com.daml.ledger.api.v1.TransactionService/GetLedgerEnd',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndResponse.FromString,
                )


class TransactionServiceServicer(object):
    """Allows clients to read transactions from the ledger.
    """

    def GetTransactions(self, request, context):
        """Read the ledger's filtered transaction stream for a set of parties.
        Lists only creates and archives, but not other events.
        Omits all events on transient contracts, i.e., contracts that were both created and archived in the same transaction.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or if the ledger has been pruned before ``begin``
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (e.g. if ``before`` is not before ``end``)
        - ``OUT_OF_RANGE``: if the ``begin`` parameter value is not before the end of the ledger
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransactionTrees(self, request, context):
        """Read the ledger's complete transaction tree stream for a set of parties.
        The stream can be filtered only by parties, but not templates (template filter must be empty).
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or if the ledger has been pruned before ``begin``
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (e.g. if ``before`` is not before ``end``)
        - ``OUT_OF_RANGE``: if the ``begin`` parameter value is not before the end of the ledger
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransactionByEventId(self, request, context):
        """Lookup a transaction tree by the ID of an event that appears within it.
        For looking up a transaction instead of a transaction tree, please see GetFlatTransactionByEventId
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or no such transaction exists
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (e.g. if requesting parties are invalid or empty)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransactionById(self, request, context):
        """Lookup a transaction tree by its ID.
        For looking up a transaction instead of a transaction tree, please see GetFlatTransactionById
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or no such transaction exists
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (e.g. if requesting parties are invalid or empty)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFlatTransactionByEventId(self, request, context):
        """Lookup a transaction by the ID of an event that appears within it.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or no such transaction exists
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (e.g. if requesting parties are invalid or empty)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFlatTransactionById(self, request, context):
        """Lookup a transaction by its ID.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or no such transaction exists
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (e.g. if requesting parties are invalid or empty)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLedgerEnd(self, request, context):
        """Get the current ledger end.
        Subscriptions started with the returned offset will serve transactions created after this RPC was called.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id or no such transaction exists
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TransactionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetTransactions': grpc.unary_stream_rpc_method_handler(
                    servicer.GetTransactions,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsResponse.SerializeToString,
            ),
            'GetTransactionTrees': grpc.unary_stream_rpc_method_handler(
                    servicer.GetTransactionTrees,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionTreesResponse.SerializeToString,
            ),
            'GetTransactionByEventId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransactionByEventId,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.SerializeToString,
            ),
            'GetTransactionById': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransactionById,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.SerializeToString,
            ),
            'GetFlatTransactionByEventId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFlatTransactionByEventId,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.SerializeToString,
            ),
            'GetFlatTransactionById': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFlatTransactionById,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.SerializeToString,
            ),
            'GetLedgerEnd': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLedgerEnd,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.TransactionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TransactionService(object):
    """Allows clients to read transactions from the ledger.
    """

    @staticmethod
    def GetTransactions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/com.daml.ledger.api.v1.TransactionService/GetTransactions',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTransactionTrees(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/com.daml.ledger.api.v1.TransactionService/GetTransactionTrees',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionTreesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTransactionByEventId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.TransactionService/GetTransactionByEventId',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTransactionById(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.TransactionService/GetTransactionById',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFlatTransactionByEventId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.TransactionService/GetFlatTransactionByEventId',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFlatTransactionById(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.TransactionService/GetFlatTransactionById',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetLedgerEnd(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.TransactionService/GetLedgerEnd',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/ledger_offset.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/ledger_offset.proto',
  package='com.daml.ledger.api.v1',
  syntax='proto3',
  serialized_options=b'\n\026com.daml.ledger.api.v1B\026LedgerOffsetOuterClass\252\002\026Com.Daml.Ledger.Api.V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n*com/daml/ledger/api/v1/ledger_offset.proto\x12\x16\x63om.daml.ledger.api.v1\"\xa8\x01\n\x0cLedgerOffset\x12\x12\n\x08\x61\x62solute\x18\x01 \x01(\tH\x00\x12G\n\x08\x62oundary\x18\x02 \x01(\x0e\x32\x33.com.daml.ledger.api.v1.LedgerOffset.LedgerBoundaryH\x00\"2\n\x0eLedgerBoundary\x12\x10\n\x0cLEDGER_BEGIN\x10\x00\x12\x0e\n\nLEDGER_END\x10\x01\x42\x07\n\x05valueBI\n\x16\x63om.daml.ledger.api.v1B\x16LedgerOffsetOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3'
)



_LEDGEROFFSET_LEDGERBOUNDARY = _descriptor.EnumDescriptor(
  name='LedgerBoundary',
  full_name='com.daml.ledger.api.v1.LedgerOffset.LedgerBoundary',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LEDGER_BEGIN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LEDGER_END', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=180,
  serialized_end=230,
)
_sym_db.RegisterEnumDescriptor(_LEDGEROFFSET_LEDGERBOUNDARY)


_LEDGEROFFSET = _descriptor.Descriptor(
  name='LedgerOffset',
  full_name='com.daml.ledger.api.v1.LedgerOffset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='absolute', full_name='com.daml.ledger.api.v1.LedgerOffset.absolute', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boundary', full_name='com.daml.ledger.api.v1.LedgerOffset.boundary', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LEDGEROFFSET_LEDGERBOUNDARY,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.LedgerOffset.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=71,
  serialized_end=239,
)

_LEDGEROFFSET.fields_by_name['boundary'].enum_type = _LEDGEROFFSET_LEDGERBOUNDARY
_LEDGEROFFSET_LEDGERBOUNDARY.containing_type = _LEDGEROFFSET
_LEDGEROFFSET.oneofs_by_name['value'].fields.append(
  _LEDGEROFFSET.fields_by_name['absolute'])
_LEDGEROFFSET.fields_by_name['absolute'].containing_oneof = _LEDGEROFFSET.oneofs_by_name['value']
_LEDGEROFFSET.oneofs_by_name['value'].fields.append(
  _LEDGEROFFSET.fields_by_name['boundary'])
_LEDGEROFFSET.fields_by_name['boundary'].containing_oneof = _LEDGEROFFSET.oneofs_by_name['value']
DESCRIPTOR.message_types_by_name['LedgerOffset'] = _LEDGEROFFSET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LedgerOffset = _reflection.GeneratedProtocolMessageType('LedgerOffset', (_message.Message,), {
  'DESCRIPTOR' : _LEDGEROFFSET,
  '__module__' : 'com.daml.ledger.api.v1.ledger_offset_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.LedgerOffset)
  })
_sym_db.RegisterMessage(LedgerOffset)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/party_management_service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/admin/party_management_service.proto',
  package='com.daml.ledger.api.v1.admin',
  syntax='proto3',
  serialized_options=b'\n\034com.daml.ledger.api.v1.adminB PartyManagementServiceOuterClass\252\002\034Com.Daml.Ledger.Api.V1.Admin',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n;com/daml/ledger/api/v1/admin/party_management_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin\"\x19\n\x17GetParticipantIdRequest\"2\n\x18GetParticipantIdResponse\x12\x16\n\x0eparticipant_id\x18\x01 \x01(\t\"$\n\x11GetPartiesRequest\x12\x0f\n\x07parties\x18\x01 \x03(\t\"W\n\x12GetPartiesResponse\x12\x41\n\rparty_details\x18\x01 \x03(\x0b\x32*.com.daml.ledger.api.v1.admin.PartyDetails\"\x19\n\x17ListKnownPartiesRequest\"]\n\x18ListKnownPartiesResponse\x12\x41\n\rparty_details\x18\x01 \x03(\x0b\x32*.com.daml.ledger.api.v1.admin.PartyDetails\"C\n\x14\x41llocatePartyRequest\x12\x15\n\rparty_id_hint\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\"Z\n\x15\x41llocatePartyResponse\x12\x41\n\rparty_details\x18\x01 \x01(\x0b\x32*.com.daml.ledger.api.v1.admin.PartyDetails\"E\n\x0cPartyDetails\x12\r\n\x05party\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x10\n\x08is_local\x18\x03 \x01(\x08\x32\x8b\x04\n\x16PartyManagementService\x12\x81\x01\n\x10GetParticipantId\x12\x35.com.daml.ledger.api.v1.admin.GetParticipantIdRequest\x1a\x36.com.daml.ledger.api.v1.admin.GetParticipantIdResponse\x12o\n\nGetParties\x12/.com.daml.ledger.api.v1.admin.GetPartiesRequest\x1a\x30.com.daml.ledger.api.v1.admin.GetPartiesResponse\x12\x81\x01\n\x10ListKnownParties\x12\x35.com.daml.ledger.api.v1.admin.ListKnownPartiesRequest\x1a\x36.com.daml.ledger.api.v1.admin.ListKnownPartiesResponse\x12x\n\rAllocateParty\x12\x32.com.daml.ledger.api.v1.admin.AllocatePartyRequest\x1a\x33.com.daml.ledger.api.v1.admin.AllocatePartyResponseB_\n\x1c\x63om.daml.ledger.api.v1.adminB PartyManagementServiceOuterClass\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3'
)




_GETPARTICIPANTIDREQUEST = _descriptor.Descriptor(
  name='GetParticipantIdRequest',
  full_name='com.daml.ledger.api.v1.admin.GetParticipantIdRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=93,
  serialized_end=118,
)


_GETPARTICIPANTIDRESPONSE = _descriptor.Descriptor(
  name='GetParticipantIdResponse',
  full_name='com.daml.ledger.api.v1.admin.GetParticipantIdResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='com.daml.ledger.api.v1.admin.GetParticipantIdResponse.participant_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=170,
)


_GETPARTIESREQUEST = _descriptor.Descriptor(
  name='GetPartiesRequest',
  full_name='com.daml.ledger.api.v1.admin.GetPartiesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parties', full_name='com.daml.ledger.api.v1.admin.GetPartiesRequest.parties', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=172,
  serialized_end=208,
)


_GETPARTIESRESPONSE = _descriptor.Descriptor(
  name='GetPartiesResponse',
  full_name='com.daml.ledger.api.v1.admin.GetPartiesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_details', full_name='com.daml.ledger.api.v1.admin.GetPartiesResponse.party_details', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=210,
  serialized_end=297,
)


_LISTKNOWNPARTIESREQUEST = _descriptor.Descriptor(
  name='ListKnownPartiesRequest',
  full_name='com.daml.ledger.api.v1.admin.ListKnownPartiesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=299,
  serialized_end=324,
)


_LISTKNOWNPARTIESRESPONSE = _descriptor.Descriptor(
  name='ListKnownPartiesResponse',
  full_name='com.daml.ledger.api.v1.admin.ListKnownPartiesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_details', full_name='com.daml.ledger.api.v1.admin.ListKnownPartiesResponse.party_details', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=326,
  serialized_end=419,
)


_ALLOCATEPARTYREQUEST = _descriptor.Descriptor(
  name='AllocatePartyRequest',
  full_name='com.daml.ledger.api.v1.admin.AllocatePartyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_id_hint', full_name='com.daml.ledger.api.v1.admin.AllocatePartyRequest.party_id_hint', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='com.daml.ledger.api.v1.admin.AllocatePartyRequest.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=421,
  serialized_end=488,
)


_ALLOCATEPARTYRESPONSE = _descriptor.Descriptor(
  name='AllocatePartyResponse',
  full_name='com.daml.ledger.api.v1.admin.AllocatePartyResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_details', full_name='com.daml.ledger.api.v1.admin.AllocatePartyResponse.party_details', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=490,
  serialized_end=580,
)


_PARTYDETAILS = _descriptor.Descriptor(
  name='PartyDetails',
  full_name='com.daml.ledger.api.v1.admin.PartyDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='party', full_name='com.daml.ledger.api.v1.admin.PartyDetails.party', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='com.daml.ledger.api.v1.admin.PartyDetails.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_local', full_name='com.daml.ledger.api.v1.admin.PartyDetails.is_local', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=582,
  serialized_end=651,
)

_GETPARTIESRESPONSE.fields_by_name['party_details'].message_type = _PARTYDETAILS
_LISTKNOWNPARTIESRESPONSE.fields_by_name['party_details'].message_type = _PARTYDETAILS
_ALLOCATEPARTYRESPONSE.fields_by_name['party_details'].message_type = _PARTYDETAILS
DESCRIPTOR.message_types_by_name['GetParticipantIdRequest'] = _GETPARTICIPANTIDREQUEST
DESCRIPTOR.message_types_by_name['GetParticipantIdResponse'] = _GETPARTICIPANTIDRESPONSE
DESCRIPTOR.message_types_by_name['GetPartiesRequest'] = _GETPARTIESREQUEST
DESCRIPTOR.message_types_by_name['GetPartiesResponse'] = _GETPARTIESRESPONSE
DESCRIPTOR.message_types_by_name['ListKnownPartiesRequest'] = _LISTKNOWNPARTIESREQUEST
DESCRIPTOR.message_types_by_name['ListKnownPartiesResponse'] = _LISTKNOWNPARTIESRESPONSE
DESCRIPTOR.message_types_by_name['AllocatePartyRequest'] = _ALLOCATEPARTYREQUEST
DESCRIPTOR.message_types_by_name['AllocatePartyResponse'] = _ALLOCATEPARTYRESPONSE
DESCRIPTOR.message_types_by_name['PartyDetails'] = _PARTYDETAILS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetParticipantIdRequest = _reflection.GeneratedProtocolMessageType('GetParticipantIdRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPARTICIPANTIDREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetParticipantIdRequest)
  })
_sym_db.RegisterMessage(GetParticipantIdRequest)

GetParticipantIdResponse = _reflection.GeneratedProtocolMessageType('GetParticipantIdResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETPARTICIPANTIDRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetParticipantIdResponse)
  })
_sym_db.RegisterMessage(GetParticipantIdResponse)

GetPartiesRequest = _reflection.GeneratedProtocolMessageType('GetPartiesRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPARTIESREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetPartiesRequest)
  })
_sym_db.RegisterMessage(GetPartiesRequest)

GetPartiesResponse = _reflection.GeneratedProtocolMessageType('GetPartiesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETPARTIESRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetPartiesResponse)
  })
_sym_db.RegisterMessage(GetPartiesResponse)

ListKnownPartiesRequest = _reflection.GeneratedProtocolMessageType('ListKnownPartiesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTKNOWNPARTIESREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPartiesRequest)
  })
_sym_db.RegisterMessage(ListKnownPartiesRequest)

ListKnownPartiesResponse = _reflection.GeneratedProtocolMessageType('ListKnownPartiesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTKNOWNPARTIESRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPartiesResponse)
  })
_sym_db.RegisterMessage(ListKnownPartiesResponse)

AllocatePartyRequest = _reflection.GeneratedProtocolMessageType('AllocatePartyRequest', (_message.Message,), {
  'DESCRIPTOR' : _ALLOCATEPARTYREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.AllocatePartyRequest)
  })
_sym_db.RegisterMessage(AllocatePartyRequest)

AllocatePartyResponse = _reflection.GeneratedProtocolMessageType('AllocatePartyResponse', (_message.Message,), {
  'DESCRIPTOR' : _ALLOCATEPARTYRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.AllocatePartyResponse)
  })
_sym_db.RegisterMessage(AllocatePartyResponse)

PartyDetails = _reflection.GeneratedProtocolMessageType('PartyDetails', (_message.Message,), {
  'DESCRIPTOR' : _PARTYDETAILS,
  '__module__' : 'com.daml.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.PartyDetails)
  })
_sym_db.RegisterMessage(PartyDetails)


DESCRIPTOR._options = None

_PARTYMANAGEMENTSERVICE = _descriptor.ServiceDescriptor(
  name='PartyManagementService',
  full_name='com.daml.ledger.api.v1.admin.PartyManagementService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=654,
  serialized_end=1177,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetParticipantId',
    full_name='com.daml.ledger.api.v1.admin.PartyManagementService.GetParticipantId',
    index=0,
    containing_service=None,
    input_type=_GETPARTICIPANTIDREQUEST,
    output_type=_GETPARTICIPANTIDRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetParties',
    full_name='com.daml.ledger.api.v1.admin.PartyManagementService.GetParties',
    index=1,
    containing_service=None,
    input_type=_GETPARTIESREQUEST,
    output_type=_GETPARTIESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListKnownParties',
    full_name='com.daml.ledger.api.v1.admin.PartyManagementService.ListKnownParties',
    index=2,
    containing_service=None,
    input_type=_LISTKNOWNPARTIESREQUEST,
    output_type=_LISTKNOWNPARTIESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AllocateParty',
    full_name='com.daml.ledger.api.v1.admin.PartyManagementService.AllocateParty',
    index=3,
    containing_service=None,
    input_type=_ALLOCATEPARTYREQUEST,
    output_type=_ALLOCATEPARTYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_PARTYMANAGEMENTSERVICE)

DESCRIPTOR.services_by_name['PartyManagementService'] = _PARTYMANAGEMENTSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This model contains definitions for the data classes used at the client layer of this library.
"""

from enum import IntEnum
from typing import NamedTuple, Optional

try:
    from typing import Literal  # type: ignore
except ImportError:
    from typing_extensions import Literal  # type: ignore


class ExitCode(IntEnum):
    """
    The different ways that a run of a dazl script can terminate.
    """

    SUCCESS = 0
    ABORT = 1
    ERROR = 2
    ERROR_INIT = 3


class LedgerRun(NamedTuple):
    """
    The results of a run of the ledger.

    Instance attributes:

    .. attribute:: LedgerRun.exit_code

        The exit code.

    .. attribute:: LedgerRun.block_start_height

        The block height of the first block consumed by the run.

    .. attribute:: LedgerRun.block_end_height

        The block height of the first block NOT consumed by the run.
    """

    exit_code: ExitCode
    block_start_height: Optional[int]
    block_end_height: Optional[int]


CREATE_IF_MISSING = Literal[1]
NONE_IF_MISSING = Literal[2]
EXCEPTION_IF_MISSING = Literal[3]

IfMissingPartyBehavior = Literal[CREATE_IF_MISSING, NONE_IF_MISSING, EXCEPTION_IF_MISSING]



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import ledger_identity_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2


class LedgerIdentityServiceStub(object):
    """Allows clients to verify that the server they are communicating with exposes the ledger they wish to operate on.
    Note that every ledger has a unique ID.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetLedgerIdentity = channel.unary_unary(
                '/com.daml.ledger.api.v1.LedgerIdentityService/GetLedgerIdentity',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityResponse.FromString,
                )


class LedgerIdentityServiceServicer(object):
    """Allows clients to verify that the server they are communicating with exposes the ledger they wish to operate on.
    Note that every ledger has a unique ID.
    """

    def GetLedgerIdentity(self, request, context):
        """Clients may call this RPC to return the identifier of the ledger they are connected to.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LedgerIdentityServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetLedgerIdentity': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLedgerIdentity,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.LedgerIdentityService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LedgerIdentityService(object):
    """Allows clients to verify that the server they are communicating with exposes the ledger they wish to operate on.
    Note that every ledger has a unique ID.
    """

    @staticmethod
    def GetLedgerIdentity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.LedgerIdentityService/GetLedgerIdentity',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import party_management_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2


class PartyManagementServiceStub(object):
    """Status: experimental interface, will change before it is deemed production
    ready

    Inspect the party management state of a ledger participant and modify the
    parts that are modifiable. We use 'backing participant' to refer to this
    specific participant in the methods of this API.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetParticipantId = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.PartyManagementService/GetParticipantId',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdResponse.FromString,
                )
        self.GetParties = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.PartyManagementService/GetParties',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetPartiesRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetPartiesResponse.FromString,
                )
        self.ListKnownParties = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.PartyManagementService/ListKnownParties',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesResponse.FromString,
                )
        self.AllocateParty = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.PartyManagementService/AllocateParty',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyResponse.FromString,
                )


class PartyManagementServiceServicer(object):
    """Status: experimental interface, will change before it is deemed production
    ready

    Inspect the party management state of a ledger participant and modify the
    parts that are modifiable. We use 'backing participant' to refer to this
    specific participant in the methods of this API.
    """

    def GetParticipantId(self, request, context):
        """Return the identifier of the backing participant.
        All horizontally scaled replicas should return the same id.
        daml-on-sql: returns an identifier supplied on command line at launch time
        daml-on-kv-ledger: as above
        canton: returns globally unique identifier of the backing participant
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetParties(self, request, context):
        """Get the party details of the given parties. Only known parties will be
        returned in the list.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListKnownParties(self, request, context):
        """List the parties known by the backing participant.
        The list returned contains parties whose ledger access is facilitated by
        backing participant and the ones maintained elsewhere.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AllocateParty(self, request, context):
        """Adds a new party to the set managed by the backing participant.
        Caller specifies a party identifier suggestion, the actual identifier
        allocated might be different and is implementation specific.
        This call may:
        - Succeed, in which case the actual allocated identifier is visible in
        the response.
        - Respond with a gRPC error
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``UNIMPLEMENTED``: if synchronous party allocation is not supported by the backing participant
        - ``INVALID_ARGUMENT``: if the provided hint and/or display name is invalid on the given ledger (see below).
        daml-on-sql: suggestion's uniqueness is checked and call rejected if the identifier is already present
        daml-on-kv-ledger: suggestion's uniqueness is checked by the validators in
        the consensus layer and call rejected if the identifier is already present.
        canton: completely different globally unique identifier is allocated.
        Behind the scenes calls to an internal protocol are made. As that protocol
        is richer than the surface protocol, the arguments take implicit values
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PartyManagementServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetParticipantId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetParticipantId,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdResponse.SerializeToString,
            ),
            'GetParties': grpc.unary_unary_rpc_method_handler(
                    servicer.GetParties,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetPartiesRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetPartiesResponse.SerializeToString,
            ),
            'ListKnownParties': grpc.unary_unary_rpc_method_handler(
                    servicer.ListKnownParties,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesResponse.SerializeToString,
            ),
            'AllocateParty': grpc.unary_unary_rpc_method_handler(
                    servicer.AllocateParty,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.admin.PartyManagementService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class PartyManagementService(object):
    """Status: experimental interface, will change before it is deemed production
    ready

    Inspect the party management state of a ledger participant and modify the
    parts that are modifiable. We use 'backing participant' to refer to this
    specific participant in the methods of this API.
    """

    @staticmethod
    def GetParticipantId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.PartyManagementService/GetParticipantId',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetParties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.PartyManagementService/GetParties',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetPartiesRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetPartiesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListKnownParties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.PartyManagementService/ListKnownParties',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AllocateParty(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.PartyManagementService/AllocateParty',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# TODO: `automodule reading` directive doesn't appear to work here, have to list each class individually.
"""
Read-Side Types
---------------

This module contains models used on the read-side of the Ledger API.

.. autoclass:: InitEvent
  :members:

.. autoclass::  InitEvent
  :members:

.. autoclass::  OffsetEvent
  :members:

.. autoclass::  ReadyEvent
  :members:

.. autoclass::  ActiveContractSetEvent
  :members:

.. autoclass::  BaseTransactionEvent
  :members:

.. autoclass::  TransactionStartEvent
  :members:

.. autoclass::  TransactionEndEvent
  :members:

.. autoclass::  ContractEvent
  :members:

.. autoclass::  ContractCreateEvent
  :members:

.. autoclass::  ContractExercisedEvent
  :members:

.. autoclass::  ContractArchiveEvent
  :members:

.. autoclass::  PackagesAddedEvent
  :members:

.. autoclass::  TransactionFilter
  :members:

.. autoclass::  EventKey
  :members:

"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Collection, Optional, Sequence, TypeVar, Union
import warnings

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.lookup import validate_template
from ..damlast.protocols import SymbolLookup
from .core import ContractContextualData, ContractData, ContractId, Party
from .lookup import template_reverse_globs

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)

    from .types import TypeReference
    from .types_store import PackageStore

T = TypeVar("T")


@dataclass(frozen=True)
class BaseEvent:
    """
    Superclass of all dazl events.
    """

    client: "Any"
    party: "Optional[Party]"
    time: "Optional[datetime]"
    ledger_id: str
    lookup: "SymbolLookup"
    package_store: "PackageStore"

    def acs_find_active(self, template: Union[TypeReference, str], match=None):
        return self.client.find_active(template, match)

    def acs_find_by_id(self, cid: Union[str, ContractId]) -> Optional[ContractContextualData]:
        return self.client.find_by_id(cid)

    def acs_find_one(self, template: Union[TypeReference, str], match=None):
        return self.client.find_one(template, match=match)

    def acs_find_historical(self, template: Union[TypeReference, str], match=None):
        return self.client.find_historical(template, match)

    def acs_find_nonempty(self, template: Union[TypeReference, str], match=None):
        return self.client.find_nonempty(template, match=match)

    def __repr__(self):
        fields = ", ".join(
            f"{k}={v!r}"
            for k, v in self.__dict__.items()
            if not k.startswith("_") and k != "client" and k != "package_store" and k != "ledger_id"
        )
        return f"{self.__class__.__name__}({fields})"


@dataclass(frozen=True)
class InitEvent(BaseEvent):
    """
    Event raised when dazl is initialized, but before it has begun reading from the Active Contract
    Set (ACS).
    """


@dataclass(frozen=True)
class OffsetEvent(BaseEvent):
    """
    Event raised when dazl is ready to begin processing new events. At this point, the Active
    Contract Set (ACS) is populated with the current state of the ledger.
    """

    offset: str


@dataclass(frozen=True)
class ReadyEvent(OffsetEvent):
    """
    Event raised when dazl is ready to begin processing new events. At this point, the Active
    Contract Set (ACS) is populated with the current state of the ledger.
    """


@dataclass(frozen=True)
class ActiveContractSetEvent(OffsetEvent):
    """
    Event raised on initial read of the active contract set.
    """

    contract_events: "Sequence[ContractCreateEvent]"


@dataclass(frozen=True)
class BaseTransactionEvent(OffsetEvent):
    """
    Event raised when dazl encounters a new transaction. This is raised before any corresponding
    :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """

    command_id: str
    workflow_id: str


@dataclass(frozen=True)
class TransactionStartEvent(BaseTransactionEvent):
    """
    Event raised when dazl encounters a new transaction. This is raised before any corresponding
    :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """

    contract_events: "Sequence[ContractEvent]"


@dataclass(frozen=True)
class TransactionEndEvent(BaseTransactionEvent):
    """
    Event raised when dazl encounters the end of a transaction. This is raised after any
    corresponding :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """

    contract_events: "Sequence[ContractEvent]"


@dataclass(frozen=True)
class ContractEvent(BaseTransactionEvent):
    """
    Event raised when dazl automation detects a new create or an archive. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """

    cid: ContractId
    cdata: ContractData
    command_id: str
    workflow_id: str
    event_id: str
    witness_parties: Sequence[str]


@dataclass(frozen=True)
class ContractCreateEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract create. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """


@dataclass(frozen=True)
class ContractExercisedEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract exercised.
    """

    contract_creating_event_id: str
    choice: str
    choice_args: Any
    acting_parties: Sequence[str]
    consuming: bool
    child_event_ids: Sequence[str]
    exercise_result: Any


@dataclass(frozen=True)
class ContractArchiveEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract archive. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """


@dataclass(frozen=True)
class PackagesAddedEvent(BaseEvent):
    """
    Event raised when new packages have been detected.
    """

    initial: bool


def create_dispatch(
    on_init: Callable[[InitEvent], T],
    on_ready: Callable[[ReadyEvent], T],
    on_offset: Callable[[OffsetEvent], T],
    on_transaction_start: Callable[[TransactionStartEvent], T],
    on_transaction_end: Callable[[TransactionEndEvent], T],
    on_contract_created: Callable[[ContractCreateEvent], T],
    on_contract_exercised: Callable[[ContractExercisedEvent], T],
    on_contract_archived: Callable[[ContractArchiveEvent], T],
    on_packages_added: Callable[[PackagesAddedEvent], T],
) -> Callable[[BaseEvent], T]:
    def handle(event: BaseEvent) -> T:
        if isinstance(event, ContractCreateEvent):
            return on_contract_created(event)
        elif isinstance(event, ContractExercisedEvent):
            return on_contract_exercised(event)
        elif isinstance(event, ContractArchiveEvent):
            return on_contract_archived(event)
        elif isinstance(event, TransactionStartEvent):
            return on_transaction_start(event)
        elif isinstance(event, TransactionEndEvent):
            return on_transaction_end(event)
        elif isinstance(event, ReadyEvent):
            return on_ready(event)
        elif isinstance(event, InitEvent):
            return on_init(event)
        elif isinstance(event, OffsetEvent):
            return on_offset(event)
        elif isinstance(event, PackagesAddedEvent):
            return on_packages_added(event)
        else:
            raise ValueError(f"unknown subclass of BaseEvent: {event!r}")

    return handle


@dataclass(frozen=True)
class ContractFilter:
    templates: "Optional[Collection[TypeConName]]" = None
    party_groups: "Optional[Collection[Party]]" = None


@dataclass(frozen=True)
class TransactionFilter(ContractFilter):
    current_offset: "Optional[str]" = None
    destination_offset: "Optional[str]" = None

    def __post_init__(self):
        if self.current_offset is not None and self.destination_offset is not None:
            if self.current_offset > self.destination_offset:
                raise ValueError(
                    "current_offset must be before destination_offset if both are specified"
                )


class EventKey:
    from_event = create_dispatch(
        on_init=lambda _: EventKey.init(),
        on_ready=lambda _: EventKey.ready(),
        on_offset=lambda _: EventKey.offset(),
        on_transaction_start=lambda _: EventKey.transaction_start(),
        on_transaction_end=lambda _: EventKey.transaction_end(),
        on_contract_created=lambda event: EventKey.contract_created(False, event.cid.value_type),
        on_contract_exercised=lambda event: EventKey.contract_exercised(
            False, event.cid.value_type, event.choice
        ),
        on_contract_archived=lambda event: EventKey.contract_archived(False, event.cid.value_type),
        on_packages_added=lambda event: EventKey.packages_added(
            initial=event.initial, changed=not event.initial
        ),
    )

    @staticmethod
    def init() -> Collection[str]:
        """
        Return the names of events that get raised in response to an :class:`InitEvent`. This is
        currently only ``'init'``.
        """
        return ("init",)

    @staticmethod
    def ready() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ReadyEvent`. This is
        currently only ``'ready'``.
        """
        return ("ready",)

    @staticmethod
    def offset() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`OffsetEvent`. This is
        currently only ``'offset'``.
        """
        return ("offset",)

    @staticmethod
    def transaction_start() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`TransactionStartEvent`.
        This is currently only ``'transaction-start'``.
        """
        return ("transaction-start",)

    @staticmethod
    def transaction_end() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`TransactionEndEvent`.
        This is currently only ``'transaction-end'``.
        """
        return ("transaction-end",)

    @staticmethod
    def contract_created(primary_only: bool, template: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractCreateEvent`
        of the specified template type.
        """
        return EventKey._contract(primary_only, "create", template)

    @staticmethod
    def contract_exercised(primary_only: bool, template: Any, choice: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractExercisedEvent`
        of the specified choice.
        """
        return [
            f"{key}/{choice}" for key in EventKey._contract(primary_only, "exercised", template)
        ]

    @staticmethod
    def contract_archived(primary_only: bool, template: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractCreateEvent`
        of the specified template type.
        """
        return EventKey._contract(primary_only, "archive", template)

    @staticmethod
    def packages_added(initial: bool, changed: bool) -> "Collection[str]":
        keys = []
        if initial:
            keys.append("packages-added/initial")
        if changed:
            keys.append("packages-added/changed")
        return tuple(keys)

    @staticmethod
    def _contract(primary_only: bool, prefix: str, template: Any) -> Collection[str]:
        m, t = validate_template(template, allow_deprecated_identifiers=True)
        return tuple(f"{prefix}/{g}" for g in template_reverse_globs(primary_only, m, t))


def max_offset(offsets: "Collection[Optional[str]]") -> "Optional[str]":
    """
    Return the most "recent" offset from a collection of offsets.

    :param offsets: A collection of offsets to examine.
    :return: The largest offset, or ``None`` if unknown.
    """
    non_none_offsets = [offset for offset in offsets if offset is not None]
    return max(non_none_offsets) if non_none_offsets else None



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the logic for fuzzy lookups (specifically, template references) are performed
in dazl.
"""
from typing import Any, Iterator, Tuple
import warnings

from ..damlast.daml_lf_1 import PackageRef


def validate_template(template: "Any") -> "Tuple[PackageRef, str]":
    warnings.warn(
        "validate_template is deprecated; use dazl.damlast.lookup.validate_template",
        DeprecationWarning,
        stacklevel=2,
    )
    from ..damlast.lookup import validate_template as validate_template_new

    if template == "*" or template is None:
        return "*", "*"

    return validate_template_new(template, allow_deprecated_identifiers=True)


def template_reverse_globs(primary_only: bool, package_id: str, type_name: str) -> "Iterator[str]":
    """
    Return an iterator over strings that glob to a specified type.
    """
    package_id = package_id or "*"
    type_name = type_name or "*"

    if package_id != "*":
        if type_name != "*":
            if ":" not in type_name:
                # this is a historical use of template name here; assume the last dot was supposed
                # to have been a colon instead
                m, delim, e = type_name.rpartition(".")
                if delim:
                    yield f"{package_id}:{m}:{e}"
                    if primary_only:
                        return
            yield f"{package_id}:{type_name}"
            if primary_only:
                return
        if not primary_only or type_name == "*":
            yield f"{package_id}:*"
    if type_name != "*":
        if ":" not in type_name:
            # this is a historical use of template name here; assume the last dot was supposed
            # to have been a colon instead
            m, delim, e = type_name.rpartition(".")
            if delim:
                yield f"*:{m}:{e}"
                if primary_only:
                    return
        yield f"*:{type_name}"
        if primary_only:
            return
    yield "*:*"



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/command_submission_service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import trace_context_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2
from . import commands_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_commands__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/command_submission_service.proto',
  package='com.daml.ledger.api.v1',
  syntax='proto3',
  serialized_options=b'\n\026com.daml.ledger.api.v1B\"CommandSubmissionServiceOuterClass\252\002\026Com.Daml.Ledger.Api.V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n7com/daml/ledger/api/v1/command_submission_service.proto\x12\x16\x63om.daml.ledger.api.v1\x1a*com/daml/ledger/api/v1/trace_context.proto\x1a%com/daml/ledger/api/v1/commands.proto\x1a\x1bgoogle/protobuf/empty.proto\"\x81\x01\n\rSubmitRequest\x12\x32\n\x08\x63ommands\x18\x01 \x01(\x0b\x32 .com.daml.ledger.api.v1.Commands\x12<\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32$.com.daml.ledger.api.v1.TraceContext2c\n\x18\x43ommandSubmissionService\x12G\n\x06Submit\x12%.com.daml.ledger.api.v1.SubmitRequest\x1a\x16.google.protobuf.EmptyBU\n\x16\x63om.daml.ledger.api.v1B\"CommandSubmissionServiceOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3'
  ,
  dependencies=[com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,com_dot_daml_dot_ledger_dot_api_dot_v1_dot_commands__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_SUBMITREQUEST = _descriptor.Descriptor(
  name='SubmitRequest',
  full_name='com.daml.ledger.api.v1.SubmitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='commands', full_name='com.daml.ledger.api.v1.SubmitRequest.commands', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.daml.ledger.api.v1.SubmitRequest.trace_context', index=1,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=196,
  serialized_end=325,
)

_SUBMITREQUEST.fields_by_name['commands'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_commands__pb2._COMMANDS
_SUBMITREQUEST.fields_by_name['trace_context'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
DESCRIPTOR.message_types_by_name['SubmitRequest'] = _SUBMITREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SubmitRequest = _reflection.GeneratedProtocolMessageType('SubmitRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.command_submission_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.SubmitRequest)
  })
_sym_db.RegisterMessage(SubmitRequest)


DESCRIPTOR._options = None

_COMMANDSUBMISSIONSERVICE = _descriptor.ServiceDescriptor(
  name='CommandSubmissionService',
  full_name='com.daml.ledger.api.v1.CommandSubmissionService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=327,
  serialized_end=426,
  methods=[
  _descriptor.MethodDescriptor(
    name='Submit',
    full_name='com.daml.ledger.api.v1.CommandSubmissionService.Submit',
    index=0,
    containing_service=None,
    input_type=_SUBMITREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_COMMANDSUBMISSIONSERVICE)

DESCRIPTOR.services_by_name['CommandSubmissionService'] = _COMMANDSUBMISSIONSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/daml_lf_dev/daml_lf_0.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/daml_lf_dev/daml_lf_0.proto',
  package='daml_lf_0',
  syntax='proto3',
  serialized_options=b'\n\024com.daml.daml_lf_dev\252\002\034Com.Daml.Daml_Lf_Dev.DamlLf0',
  serialized_pb=b'\n$com/daml/daml_lf_dev/daml_lf_0.proto\x12\tdaml_lf_0\"\t\n\x07PackageB5\n\x14\x63om.daml.daml_lf_dev\xaa\x02\x1c\x43om.Daml.Daml_Lf_Dev.DamlLf0b\x06proto3'
)




_PACKAGE = _descriptor.Descriptor(
  name='Package',
  full_name='daml_lf_0.Package',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=51,
  serialized_end=60,
)

DESCRIPTOR.message_types_by_name['Package'] = _PACKAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Package = _reflection.GeneratedProtocolMessageType('Package', (_message.Message,), {
  'DESCRIPTOR' : _PACKAGE,
  '__module__' : 'com.daml.daml_lf_dev.daml_lf_0_pb2'
  # @@protoc_insertion_point(class_scope:daml_lf_0.Package)
  })
_sym_db.RegisterMessage(Package)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/daml_lf_dev/daml_lf.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import daml_lf_1_pb2 as com_dot_daml_dot_daml__lf__dev_dot_daml__lf__1__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/daml_lf_dev/daml_lf.proto',
  package='daml_lf_dev',
  syntax='proto3',
  serialized_options=b'\n\024com.daml.daml_lf_dev\252\002\033Com.Daml.Daml_Lf_Dev.DamlLf',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"com/daml/daml_lf_dev/daml_lf.proto\x12\x0b\x64\x61ml_lf_dev\x1a$com/daml/daml_lf_dev/daml_lf_1.proto\"]\n\x0e\x41rchivePayload\x12\r\n\x05minor\x18\x03 \x01(\t\x12\'\n\tdaml_lf_1\x18\x02 \x01(\x0b\x32\x12.daml_lf_1.PackageH\x00\x42\x05\n\x03SumJ\x06\x08\x8fN\x10\x90NJ\x04\x08\x01\x10\x02\"Z\n\x07\x41rchive\x12\x30\n\rhash_function\x18\x01 \x01(\x0e\x32\x19.daml_lf_dev.HashFunction\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x12\x0c\n\x04hash\x18\x04 \x01(\t*\x1a\n\x0cHashFunction\x12\n\n\x06SHA256\x10\x00\x42\x34\n\x14\x63om.daml.daml_lf_dev\xaa\x02\x1b\x43om.Daml.Daml_Lf_Dev.DamlLfb\x06proto3'
  ,
  dependencies=[com_dot_daml_dot_daml__lf__dev_dot_daml__lf__1__pb2.DESCRIPTOR,])

_HASHFUNCTION = _descriptor.EnumDescriptor(
  name='HashFunction',
  full_name='daml_lf_dev.HashFunction',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SHA256', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=276,
  serialized_end=302,
)
_sym_db.RegisterEnumDescriptor(_HASHFUNCTION)

HashFunction = enum_type_wrapper.EnumTypeWrapper(_HASHFUNCTION)
SHA256 = 0



_ARCHIVEPAYLOAD = _descriptor.Descriptor(
  name='ArchivePayload',
  full_name='daml_lf_dev.ArchivePayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='minor', full_name='daml_lf_dev.ArchivePayload.minor', index=0,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='daml_lf_1', full_name='daml_lf_dev.ArchivePayload.daml_lf_1', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Sum', full_name='daml_lf_dev.ArchivePayload.Sum',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=89,
  serialized_end=182,
)


_ARCHIVE = _descriptor.Descriptor(
  name='Archive',
  full_name='daml_lf_dev.Archive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_function', full_name='daml_lf_dev.Archive.hash_function', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='daml_lf_dev.Archive.payload', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hash', full_name='daml_lf_dev.Archive.hash', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=184,
  serialized_end=274,
)

_ARCHIVEPAYLOAD.fields_by_name['daml_lf_1'].message_type = com_dot_daml_dot_daml__lf__dev_dot_daml__lf__1__pb2._PACKAGE
_ARCHIVEPAYLOAD.oneofs_by_name['Sum'].fields.append(
  _ARCHIVEPAYLOAD.fields_by_name['daml_lf_1'])
_ARCHIVEPAYLOAD.fields_by_name['daml_lf_1'].containing_oneof = _ARCHIVEPAYLOAD.oneofs_by_name['Sum']
_ARCHIVE.fields_by_name['hash_function'].enum_type = _HASHFUNCTION
DESCRIPTOR.message_types_by_name['ArchivePayload'] = _ARCHIVEPAYLOAD
DESCRIPTOR.message_types_by_name['Archive'] = _ARCHIVE
DESCRIPTOR.enum_types_by_name['HashFunction'] = _HASHFUNCTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ArchivePayload = _reflection.GeneratedProtocolMessageType('ArchivePayload', (_message.Message,), {
  'DESCRIPTOR' : _ARCHIVEPAYLOAD,
  '__module__' : 'com.daml.daml_lf_dev.daml_lf_pb2'
  # @@protoc_insertion_point(class_scope:daml_lf_dev.ArchivePayload)
  })
_sym_db.RegisterMessage(ArchivePayload)

Archive = _reflection.GeneratedProtocolMessageType('Archive', (_message.Message,), {
  'DESCRIPTOR' : _ARCHIVE,
  '__module__' : 'com.daml.daml_lf_dev.daml_lf_pb2'
  # @@protoc_insertion_point(class_scope:daml_lf_dev.Archive)
  })
_sym_db.RegisterMessage(Archive)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import active_contracts_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2


class ActiveContractsServiceStub(object):
    """Allows clients to initialize themselves according to a fairly recent state of the ledger without reading through all transactions that were committed since the ledger's creation.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetActiveContracts = channel.unary_stream(
                '/com.daml.ledger.api.v1.ActiveContractsService/GetActiveContracts',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsResponse.FromString,
                )


class ActiveContractsServiceServicer(object):
    """Allows clients to initialize themselves according to a fairly recent state of the ledger without reading through all transactions that were committed since the ledger's creation.
    """

    def GetActiveContracts(self, request, context):
        """Returns a stream of the latest snapshot of active contracts.
        If there are no active contracts, the stream returns a single GetActiveContractsResponse message with the offset at which the snapshot has been taken.
        Clients SHOULD use the offset in the last GetActiveContractsResponse message to continue streaming transactions with the transaction service.
        Clients SHOULD NOT assume that the set of active contracts they receive reflects the state at the ledger end.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields (filters by party cannot be empty)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ActiveContractsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetActiveContracts': grpc.unary_stream_rpc_method_handler(
                    servicer.GetActiveContracts,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.ActiveContractsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ActiveContractsService(object):
    """Allows clients to initialize themselves according to a fairly recent state of the ledger without reading through all transactions that were committed since the ledger's creation.
    """

    @staticmethod
    def GetActiveContracts(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/com.daml.ledger.api.v1.ActiveContractsService/GetActiveContracts',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import datetime
from threading import Event, RLock, Thread
from typing import Awaitable, Dict, Iterable, Optional, Union

from .. import LOG
from ..model.core import ConnectionTimeoutError, UserTerminateRequest
from ..model.ledger import LedgerMetadata
from ..model.network import HTTPConnectionSettings
from ..prim import Party
from ..scheduler import Invoker
from ._base import LedgerClient, LedgerConnectionOptions, LedgerNetwork
from .oauth import oauth_flow
from .v1.grpc import GRPCv1Connection


class AutodetectLedgerNetwork(LedgerNetwork):
    """
    Auto-detecting pool implementation that automatically routes to the correct pool
    implementation based on the scheme specified in the URL.
    """

    def __init__(self, invoker: "Invoker", options: "LedgerConnectionOptions"):
        self._invoker = invoker
        self._options = options

        self._closed = False
        self._first_connection_evt = Event()
        self._first_connection = None  # type: Optional[AutodetectConnection]
        self._connections = dict()  # type: Dict[HTTPConnectionSettings, AutodetectConnection]
        self._lock = RLock()
        self._ledger_future = self._invoker.create_future()
        self._main_thread = Thread(target=self._main, daemon=True)
        self._main_thread.start()

    def ledger(self) -> "Awaitable[LedgerMetadata]":
        return self._ledger_future

    async def connect_anonymous(
        self, settings: "HTTPConnectionSettings", context_path: "Optional[str]"
    ) -> None:
        """
        Establish a single no-Party connection (but only if no other connections have already been
        established). This is used by specialized setups that do not require Parties to be supplied
        for any reason (such as fetching initial ledger metadata).
        """
        self._get_connection(settings, context_path)
        await self.ledger()

    async def connect(
        self,
        party: "Union[str, Party]",
        settings: "HTTPConnectionSettings",
        context_path: "Optional[str]" = None,
    ) -> LedgerClient:
        LOG.info("Establishing a connection to %s on party %s...", settings, party)
        if settings.oauth:
            new_oauth_settings = await oauth_flow(settings.oauth)
            settings = settings._replace(oauth=new_oauth_settings)
        conn = self._get_connection(settings, context_path)
        ledger = await self.ledger()
        if ledger.protocol_version == "v1":
            from .v1.grpc import GRPCv1LedgerClient

            return GRPCv1LedgerClient(conn, ledger, party)
        elif ledger.protocol_version == "v0":
            raise RuntimeError(f"Unsupported protocol version: {ledger.protocol_version}")
        else:
            raise RuntimeError(f"Unknown protocol version: {ledger.protocol_version}")

    async def upload_package(self, dar_contents: bytes) -> None:
        from .v1.grpc import grpc_upload_package

        return await self._invoker.run_in_executor(
            lambda: grpc_upload_package(self._first_connection, dar_contents)
        )

    async def set_time(self, new_time: datetime):
        ledger = await self.ledger()
        if ledger.protocol_version == "v1":
            from .v1.grpc import grpc_set_time

            return await self._invoker.run_in_executor(
                lambda: grpc_set_time(self._first_connection, ledger.ledger_id, new_time)
            )
        elif ledger.protocol_version == "v0":
            raise RuntimeError(f"Unsupported protocol version: {ledger.protocol_version}")
        else:
            raise RuntimeError(f"Unknown protocol version: {ledger.protocol_version}")

    async def close(self) -> None:
        with self._lock:
            connections = list(self._connections.values())
            self._connections.clear()

        for connection in connections:
            try:
                connection.close()
            except:
                LOG.exception("Had trouble closing a connection.")

    def _close_all(self) -> None:
        LOG.debug("Closing all connections...")
        try:
            with self._lock:
                connections = list(self._connections.values())
                self._connections.clear()

            for connection in connections:
                # noinspection PyBroadException
                try:
                    connection.close()
                except Exception:  # noqa
                    LOG.exception("Had trouble closing a connection.")
        finally:
            LOG.debug("Marked the connection pool as closed.")
            self._closed = True

    @property
    def closed(self):
        return self._closed

    def _get_connection(
        self, settings: HTTPConnectionSettings, context_path: Optional[str]
    ) -> "AutodetectConnection":

        if not self._lock.acquire(timeout=5):
            raise TimeoutError("Could not acquire an internal lock quickly enough.")
        try:
            if self._closed:
                raise IOError("LedgerNetwork.close() has already been called")

            if not self._connections:
                LOG.debug("Initializing the first connection to %s...", settings)
                self._first_connection = AutodetectConnection(
                    self._invoker, self._options, settings, context_path
                )
                self._connections[settings] = self._first_connection
                self._first_connection_evt.set()
                return self._first_connection
            else:
                LOG.debug("Initializing a connection to %s...", settings)
                stub = self._connections.get(settings)
                if stub is None:
                    stub = AutodetectConnection(self._invoker, settings, context_path)
                    self._connections[settings] = stub
                return stub
        except:  # noqa
            LOG.exception("An error occurred trying to create a connection.")
            raise
        finally:
            self._lock.release()

    def _main(self) -> None:
        timeout = self._options.connect_timeout
        try:
            LOG.debug(
                "Waiting for the first connection to be established (timeout: %s)...", timeout
            )
            if not self._first_connection_evt.wait(
                timeout=timeout.total_seconds() if timeout is not None else None
            ):
                LOG.error(
                    "Waited %s for the first connection but it never came. Aborting...", timeout
                )
                raise Exception("first connection timeout")

            LOG.debug("Starting ledger detection.")
            conn = self._first_connection
            if conn is not None:
                for i, metadata in enumerate(_monitor_ledger_network(conn)):
                    if i == 0:
                        self._invoker.run_in_loop(lambda: self._ledger_future.set_result(metadata))
                    elif metadata is not None:
                        LOG.warning("The network monitor thread emitted multiple metadata!")
        except (UserTerminateRequest, ConnectionTimeoutError) as ex:
            # re-raise these, but they are "known" errors so a stack trace in the logs would just
            # create clutter
            self._invoker.run_in_loop(lambda: self._ledger_future.set_exception(ex))

        except Exception as ex:
            # unexpected exception raised, so provide lots of information that might help debug
            LOG.exception("The main monitoring thread died.")

            def _maybe_apply_error():
                if not self._ledger_future.done():
                    LOG.error("The above error was propagated as an initialization error.")
                    self._ledger_future.set_exception(ex)

            # if we haven't even been able to initialize, propagate this error to the init future
            self._invoker.run_in_loop(_maybe_apply_error)

        finally:
            self._close_all()


class AutodetectConnection(GRPCv1Connection):
    pass


def _monitor_ledger_network(connection: AutodetectConnection) -> Iterable[LedgerMetadata]:
    """
    Monitor the very first connection established and provide general information about
    the ledger to other parties.
    """
    from .v1.grpc import grpc_detect_ledger_id

    ledger_id = grpc_detect_ledger_id(connection)
    if ledger_id is not None:
        LOG.info("Ledger ID: %s", ledger_id)
        from .v1.grpc import grpc_main_thread

        return grpc_main_thread(connection=connection, ledger_id=ledger_id)
    elif connection.settings.scheme in ("grpc", "grpcs"):
        raise Exception("The protocol was specified as gRPC, but the backend does not support it.")
    else:
        LOG.error("Couldn't connect over gRPC.")
        raise Exception("Connection to the v0 REST API is no longer supported.")



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Sequence

from ..damlast.daml_lf_1 import Expr, FieldWithType
from .types import ModuleRef


class DamlDeclaration:
    """"""

    module: ModuleRef
    name: Sequence[str]


class DamlDataType(DamlDeclaration):
    """"""

    module: ModuleRef
    name: Sequence[str]
    fields: "Sequence[FieldWithType]"
    eq_instance: "Optional[Expr]"
    ord_instance: "Optional[Expr]"
    show_instance: "Optional[Expr]"


class DamlTemplate(DamlDataType):
    signatories: "Expr"
    observers: "Expr"
    agreement: "Expr"
    precondition: "Expr"
    choices: "Sequence[DamlChoice]"


class DamlChoice:
    name: str



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from pygments.lexer import inherit
from pygments.lexers.haskell import HaskellLexer
from pygments.token import Keyword


class DAMLLexer(HaskellLexer):

    name = "DAML"
    aliases = ["daml"]
    filenames = ["*.daml"]

    daml_reserved = (
        "template",
        "with",
        "controller",
        "can",
        "ensure",
        "daml",
        "observer",
        "signatory",
        "agreement",
        "controller",
        "nonconsuming",
        "return",
        "this",
    )

    tokens = {
        "root": [
            (r"\b(%s)(?!\')\b" % "|".join(daml_reserved), Keyword.Reserved),
            (r"\b(True|False)\b", Keyword.Constant),
            inherit,
        ]
    }



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
import json
from typing import Callable, Dict, Optional, Sequence, Type as TType, Union

from .. import LOG
from ..damlast.daml_lf_1 import (
    Block,
    BuiltinFunction,
    Case,
    CaseAlt,
    DefDataType,
    DefTemplate,
    DefValue,
    Expr,
    Module,
    ModuleRef,
    Package,
    PrimCon,
    PrimLit,
    Scenario,
    Type,
    TypeVarWithKind,
    Unit,
    Update,
    ValName,
)
from ..damlast.protocols import SymbolLookup
from ..damlast.util import package_local_name, unpack_arrow_type
from ..damlast.visitor import ExprVisitor, ModuleVisitor, PackageVisitor, TypeVisitor
from ..model.definition import DamlDataType, DamlTemplate
from ..prim import to_date, to_datetime
from .options import PrettyOptions
from .util import is_hidden_module_name, maybe_parentheses


# noinspection PyMethodMayBeStatic
class PrettyPrintBase(PackageVisitor[str], ModuleVisitor[str], ExprVisitor[str], TypeVisitor[str]):
    """
    Convenience base class for shared code between all pretty-print implementations.
    """

    def __init__(
        self,
        lookup: "Optional[SymbolLookup]" = None,
        context: "Union[None, CodeContext, PrettyOptions]" = None,
    ):
        self.lookup = lookup
        if context is None:
            self.context = CodeContext.top_level()
        elif isinstance(context, CodeContext):
            self.context = context
        elif isinstance(context, PrettyOptions):
            self.context = CodeContext.from_options(context)
        else:
            raise TypeError("a CodeContext object required here")

    def lexer(self):
        """
        Return a Pygments lexer that can be used to render the results from this pretty printer,
        or :class:`None` if none are defined or Pygments could not be loaded.
        """
        return None

    def module_indent(self):
        from ..damlast.util import module_name

        cm = self.context.current_module
        if cm is not None:
            mn = module_name(cm)
            indent_level = str(mn).count(".") if mn else 0
            return " " * (indent_level * 4)
        else:
            return ""

    def render_store(self) -> str:
        """
        Render everything in a :class:`PackageStore`.
        """
        with StringIO() as buf:
            buf.write("from dazl import create, exercise, module, TemplateMeta, ChoiceMeta\n\n")
            for archive in self.lookup.archives():
                buf.write(self.visit_package(archive.package))
                buf.write("\n")
            return buf.getvalue()

    def with_module(self, module_ref: "ModuleRef"):
        return type(self)(lookup=self.lookup, context=self.context.with_module(module_ref))

    def with_decl(self, decl_name: Sequence[str], decl_type: Type):
        """
        Return a sub-scoped pretty print visitor under the context of a field with a specified name
        and type.

        :param decl_name: The name of the declaration.
        :param decl_type: The type of the declaration.
        :return: An instance of this type.
        """
        return type(self)(lookup=self.lookup, context=self.context.with_decl(decl_name, decl_type))

    def with_type_abs(self, type_abs: "Sequence[TypeVarWithKind]"):
        return type(self)(lookup=self.lookup, context=self.context.with_type_abs(type_abs))

    def with_type_app(self, type_app: "Sequence[Type]"):
        return type(self)(lookup=self.lookup, context=self.context.with_type_app(type_app))

    def with_expression(self):
        return type(self)(lookup=self.lookup, context=self.context.with_expression())

    def with_statement_block(self):
        return type(self)(lookup=self.lookup, context=self.context.with_statement_block())

    def visit_package(self, package: "Package") -> "str":
        """
        Simply render each package as the contents of its constituent models.

        :param package:
            The :class:`Package` of :class:`Module` instances to render.
        :return:
            The string representation of the package, as defined by the string representation of
            each of its modules.
        """
        LOG.info("Printing a package with %d modules:", len(package.modules))
        from ._module_builder import ModuleHierarchy

        module_map = ModuleHierarchy("")
        for module in package.modules:
            if self.context.show_hidden_types or not is_hidden_module_name(module.name.segments):
                module_map.add_module(module)

        lines = []
        for event_kind, ref, module in module_map:
            if event_kind == ModuleHierarchy.START:
                lines.append(self.visit_module_ref_start(ref))
            elif event_kind == ModuleHierarchy.ITEM:
                child = self.with_module(ref)
                indent = child.module_indent()
                module_contents = "\n".join(
                    indent + line for line in child.visit_module(module).splitlines()
                )
                lines.append(module_contents or (indent + "pass"))
            elif event_kind == ModuleHierarchy.END:
                lines.append(self.visit_module_ref_end(ref))

        return "\n".join(lines)

    def visit_module(self, module: "Module") -> "str":
        """
        Attempt to render the contents of a module.

        :param module: The module to render.
        :return: A string that represents the full contents of this module.
        """
        at_least_one = False

        lines = []
        # render raw values
        for value in module.values:
            if self.context.show_hidden_types or not value.name_with_type.name[0].startswith("$"):
                at_least_one = True
                lines.append(self.visit_def_value(value) + "\n\n")

        # now render data (and templates)
        data_types_by_name = {".".join(dt.name.segments): dt for dt in module.data_types}
        templates_by_name = {".".join(t.tycon.segments): t for t in module.templates}

        bare_data_types = {}
        template_data_types = {}
        for key in sorted({*data_types_by_name.keys(), *templates_by_name.keys()}):
            data_type = data_types_by_name.get(key)
            template = templates_by_name.get(key)
            if template is None:
                bare_data_types[key] = (None, data_type)
            else:
                template_data_types[key] = (template, data_type)

        for template, data_type in (*bare_data_types.values(), *template_data_types.values()):
            if self.context.show_hidden_types or not data_type.name.segments[0].startswith("$"):
                at_least_one = True
                lines.append(self.visit_def_template(template, data_type))
                lines.append("")

        if not at_least_one:
            self.visit_empty_block_body()

        return "\n".join(lines)

    def visit_empty_block_body(self):
        return ""

    def visit_module_ref_start(self, module_ref: "ModuleRef") -> "str":
        """
        Render the "start" of a module.
        """
        return ""

    def visit_module_ref_end(self, module_ref: "ModuleRef") -> "str":
        """
        Render the "end" of a module.
        """
        return ""

    def visit_def_data_type(self, def_data_type: "DefDataType") -> str:
        return self.visit_def_template(None, def_data_type)

    def visit_def_template(
        self, template: "DefTemplate", def_data_type: "Optional[DefDataType]" = None
    ) -> str:
        """
        Attempt to render the :class:`DefTemplate`. If this method successfully returns, then the
        corresponding :class:`DefValue` and :class:`DefDataType` objects are skipped.

        :param template: The :class:`DefTemplate` to attempt to render.
        :param def_data_type: The corresponding :class:`DefDataType` for that template (if known).
        :return: A rendered version of the template.
        """
        raise NotImplementedError

    def visit_daml_data_type(self, data_type: DamlDataType) -> str:
        """
        Render the specified data type.

        :param data_type:
        :return:
        """
        raise NotImplementedError

    def visit_daml_template(self, template: DamlTemplate) -> str:
        """
        Render a template instance.
        """
        raise NotImplementedError

    def visit_def_value(self, def_value: "DefValue") -> str:
        scope = self.with_decl(def_value.name_with_type.name, def_value.name_with_type.type)
        return scope.visit_expr(def_value.expr)

    def visit_expr(self, expr: "Expr") -> "str":
        try:
            expr_text = super(PrettyPrintBase, self).visit_expr(expr)
            if expr_text is None:
                LOG.warning(
                    "All visit_expr_* methods should return something (expr_%s did not)",
                    expr._Sum_name,
                )
                return self._visit_expr_default()
            else:
                return expr_text
        except:
            LOG.exception("Could not render an expression!")
            return self._visit_expr_default()

    def visit_expr_var(self, var: str) -> "str":
        """
        Render the variable. The default implementation merely returns the variable, unchanged.

        :param var: The variable to render.
        :return: The rendered value.
        """
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_var_decl(var, self.context)
        else:
            return self.visit_expr_var_inline(var)

    def visit_expr_var_decl(self, var: str, context: "CodeContext") -> "str":
        return self._visit_expr_decl(context, self.visit_expr_var_inline(var))

    def visit_expr_var_inline(self, var: str) -> "str":
        return var

    def visit_expr_val(self, val: "ValName") -> "str":
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_val_decl(val, self.context)
        else:
            return self.visit_expr_val_inline(val)

    def visit_expr_val_decl(self, val: "ValName", context: "CodeContext") -> "str":
        return self._visit_expr_decl(context, self.visit_expr_val_inline(val))

    def visit_expr_val_inline(self, val: "ValName"):
        return package_local_name(val)

    def visit_expr_builtin(self, builtin: "BuiltinFunction") -> "str":
        return builtin.name

    def visit_expr_prim_con(self, prim_con: "PrimCon") -> "str":
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_prim_con_decl(prim_con, self.context)
        else:
            return self.visit_expr_prim_con_inline(prim_con)

    def visit_expr_prim_con_decl(self, prim_con: "PrimCon", context: "CodeContext") -> str:
        return self._visit_expr_decl(context, self.visit_expr_prim_con_inline(prim_con))

    def visit_expr_prim_con_inline(self, prim_con: "PrimCon") -> "str":
        if PrimCon.CON_UNIT == prim_con:
            return self.visit_expr_prim_con_inline_unit()
        elif PrimCon.CON_TRUE == prim_con:
            return self.visit_expr_prim_con_inline_true()
        elif PrimCon.CON_FALSE == prim_con:
            return self.visit_expr_prim_con_inline_false()
        else:
            return repr(prim_con)

    def visit_expr_prim_con_inline_true(self) -> "str":
        return "True"

    def visit_expr_prim_con_inline_false(self) -> "str":
        return "False"

    def visit_expr_prim_con_inline_unit(self) -> "str":
        return "Unit"

    def visit_expr_prim_lit(self, prim_lit: "PrimLit") -> "str":
        """
        Visit a primitive literal.
        """
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_prim_lit_decl(prim_lit, self.context)
        else:
            return self.visit_expr_prim_lit_inline(prim_lit)

    def visit_expr_prim_lit_decl(self, prim_lit: "PrimLit", context: "CodeContext") -> str:
        return self._visit_expr_decl(context, self.visit_expr_prim_lit_inline(prim_lit))

    def visit_expr_prim_lit_inline(self, prim_lit: "PrimLit") -> str:
        if prim_lit.text is not None:
            return self.visit_expr_prim_lit_inline_text(prim_lit.text)
        elif prim_lit.int64 is not None:
            return self.visit_expr_prim_lit_inline_int64(prim_lit.int64)
        elif prim_lit.party is not None:
            return self.visit_expr_prim_lit_inline_party(prim_lit.party)
        elif prim_lit.decimal is not None:
            return self.visit_expr_prim_lit_inline_decimal(prim_lit.decimal)
        elif prim_lit.date is not None:
            return self.visit_expr_prim_lit_inline_date(prim_lit.date)
        elif prim_lit.timestamp is not None:
            return self.visit_expr_prim_lit_inline_timestamp(prim_lit.timestamp)
        else:
            return repr(prim_lit)

    def visit_expr_prim_lit_inline_decimal(self, decimal: str) -> str:
        """
        Visit a primitive decimal literal. The default implementation merely returns the original
        string that is the specified value, which tends to be a good default in most languages.

        :param decimal: The numeric value to format.
        :return: The text value.
        """
        return decimal

    def visit_expr_prim_lit_inline_int64(self, int64: int) -> str:
        """
        Visit a primitive text literal. The default implementation uses the JSON encoding of the
        specified value, which tends to be a good default in most languages.

        :param int64: The numeric value to format.
        :return: The text value.
        """
        return json.dumps(int64)

    def visit_expr_prim_lit_inline_text(self, text: str) -> str:
        """
        Visit a primitive text literal. The default implementation uses the JSON encoding of the
        specified value, which tends to be a good default in most languages.

        :param text: The text value to format.
        :return: The text value.
        """
        return json.dumps(text)

    def visit_expr_prim_lit_inline_party(self, party: str) -> str:
        """
        Visit a primitive party literal. The default implementation uses the Python repr encoding,
        which formats strings with a single quote.

        :param party: The text value to format.
        :return: The text value, with quotes escaped.
        """
        return repr(party)

    def visit_expr_prim_lit_inline_date(self, date: int) -> str:
        return to_date(date).isoformat()

    def visit_expr_prim_lit_inline_timestamp(self, timestamp: float) -> str:
        return to_datetime(timestamp).isoformat()

    def visit_expr_rec_con(self, rec_con: "Expr.RecCon") -> "str":
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_rec_con_decl(rec_con, self.context)
        else:
            return self.visit_expr_rec_con_inline(rec_con)

    def visit_expr_rec_con_decl(self, rec_con: "Expr.RecCon", context: "CodeContext") -> "str":
        expr_text = self.visit_expr_rec_con_inline(rec_con)
        return self._visit_expr_decl(context, expr_text)

    def visit_expr_rec_con_inline(self, rec_con: "Expr.RecCon") -> "str":
        raise NotImplementedError

    def visit_expr_rec_proj(self, rec_proj: "Expr.RecProj") -> "str":
        pass

    def visit_expr_enum_con(self, enum_con: "Expr.EnumCon") -> "str":
        return enum_con.enum_con

    def visit_expr_variant_con(self, variant_con: "Expr.VariantCon") -> "str":
        pass

    def visit_expr_struct_con(self, struct_con: "Expr.StructCon") -> "str":
        pass

    def visit_expr_struct_proj(self, struct_proj: "Expr.StructProj") -> "str":
        pass

    def visit_expr_app(self, app: "Expr.App") -> "str":
        """
        Visit an application (function called with one or more arguments).
        """
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_app_decl(app, self.context)
        else:
            return self.visit_expr_app_inline(app)

    def visit_expr_app_decl(self, app: "Expr.App", context: "CodeContext"):
        """
        Visit the declaration of a field whose value is the application of a function with one or
        more parameters. The default implementation assumes that the declaration is identical to the
        inline case.

        :param app:
            The :class:`Expr.App` that is the value of the declaration.
        :param context:
            The _enclosing_ (not the current) context, which includes information about the name and
            type of the declaration.
        :return:
            A string that represents the definition of a value with the name specified by context.
        """
        expr_text = self.visit_expr_app_inline(app)
        return self._visit_expr_decl(context, expr_text)

    def visit_expr_app_inline(self, app: "Expr.App") -> str:
        """
        Visit an application used in the middle of an expression.
        """
        raise NotImplementedError("visit_expr_app_inline needs an implementation")

    def visit_expr_ty_app(self, ty_app: "Expr.TyApp") -> "str":
        """
        Visit the application of a type parameter with a body parameterized on type.

        The default implementation merely creates a child context with type arguments applied, and
        renders the body of the expression at that child context.

        :param ty_app:
            The :class:`Expr.TypeApp` to render.
        :return:
            A string that represents the :class:`Expr.TypeApp`.
        """
        return self.with_type_app(ty_app.types).visit_expr(ty_app.expr)

    def visit_expr_abs(self, abs_: "Expr.Abs") -> "str":
        """
        Visit an abstraction (function).
        """
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_abs_decl(abs_, self.context)
        else:
            return self.visit_expr_abs_inline(abs_)

    def visit_expr_abs_decl(self, abs_: "Expr.Abs", context: "CodeContext"):
        """
        Visit the declaration of an abstraction (function). The default implementation assumes that
        function declarations are identical to the inline case (which is not normally the case in
        most languages, but many support this as an alternate syntax).

        :param abs_:
            The :class:`Expr.Abs` that is the value of the declaration.
        :param context:
            The _enclosing_ (not the current) context, which includes information about the name and
            type of the declaration.
        :return:
            A string that represents the definition of a value with the name specified by context.
        """
        expr_text = self.visit_expr_abs_inline(abs_)
        return self._visit_expr_decl(context, expr_text)

    def visit_expr_abs_decl_eq(self, abs_: "Expr.Abs", context: "CodeContext"):
        """
        Visit the declaration of an abstraction (function) that is the implementation of the
        equality check in the ``GHC.Classes.Eq`` typeclass.

        This is special-cased because in almost all languages, special syntax is required to
        overload equality.

        :param abs_:
            The :class:`Expr.Abs` that is the value of the declaration.
        :param context:
            The _enclosing_ (not the current) context, which includes information about the name and
            type of the declaration.
        :return:
            A string that represents the definition of a value with the name specified by context.
        """
        expr_text = self.visit_expr_abs_inline(abs_)
        return self._visit_expr_decl(context, expr_text)

    def visit_expr_abs_decl_ne(self, abs_: "Expr.Abs", context: "CodeContext"):
        """
        Visit the declaration of an abstraction (function) that is the implementation of the
        equality check in the ``GHC.Classes.Eq`` typeclass.

        This is special-cased because in almost all languages, special syntax is required to
        overload inequality.

        :param abs_:
            The :class:`Expr.Abs` that is the value of the declaration.
        :param context:
            The _enclosing_ (not the current) context, which includes information about the name and
            type of the declaration.
        :return:
            A string that represents the definition of a value with the name specified by context.
        """
        expr_text = self.visit_expr_abs_inline(abs_)
        return self._visit_expr_decl(context, expr_text)

    def visit_expr_abs_inline(self, abs_: "Expr.Abs") -> str:
        """
        Visit an inline function used in the middle of an expression.
        """
        raise NotImplementedError("visit_expr_abs_inline needs an implementation")

    def visit_expr_ty_abs(self, ty_abs: "Expr.TyAbs") -> "str":
        """
        Visit the abstraction of a type parameter with a body parameterized on type.

        The default implementation merely creates a child context with type arguments applied, and
        renders the body of the expression at that child context.

        :param ty_abs:
            The :class:`Expr.TypeAbs` to render.
        :return:
            A string that represents the :class:`Expr.TypeAbs`.
        """
        return self.with_type_abs(ty_abs.param).visit_expr(ty_abs.body)

    def visit_expr_case(self, case: "Case") -> "str":
        return None

    def visit_expr_casealt(self, alt: "CaseAlt", type: "Optional[Type]" = None) -> str:
        pattern_text = alt.Sum_match(
            self.visit_expr_casealt_default,
            self.visit_expr_casealt_variant,
            self.visit_expr_casealt_prim_con,
            self.visit_expr_casealt_nil,
            self.visit_expr_casealt_cons,
            self.visit_expr_casealt_optional_none,
            self.visit_expr_casealt_optional_some,
            self.visit_expr_casealt_enum,
        )
        body_text = self.visit_expr(alt.body)
        return self._visit_expr_casealt(pattern_text, body_text)

    def _visit_expr_casealt(self, pattern_text: str, body_text: str) -> str:
        return f"{pattern_text} -> {maybe_parentheses(body_text)}"

    def visit_expr_casealt_default(self, default: "Unit"):
        return "_"

    def visit_expr_casealt_variant(self, variant: "CaseAlt.Variant"):
        var_con = Expr.VariantCon(
            # TODO: Probably not correct
            tycon=Type.Con(tycon=variant.con, args=()),
            variant_con=variant.variant,
            variant_arg=Expr(var=variant.binder),
        )
        return self.visit_expr_variant_con(var_con)

    def visit_expr_casealt_prim_con(self, prim_con: "PrimCon"):
        return self.visit_expr_prim_con(prim_con)

    def visit_expr_casealt_nil(self, nil: "Unit", type: "Optional[Type]" = None):
        return self.visit_expr_nil(Expr.Nil(type))

    def visit_expr_casealt_cons(self, cons: "CaseAlt.Cons", type: "Optional[Type]" = None):
        return self.visit_expr_cons(
            Expr.Cons(front=(Expr(var=cons.var_head),), tail=Expr(var=cons.var_tail), type=type)
        )

    def visit_expr_casealt_optional_none(
        self, optional_none: "Unit", type: "Optional[Type]" = None
    ):
        return self.visit_expr_optional_none(Expr.OptionalNone(type=type))

    def visit_expr_casealt_optional_some(
        self, optional_some: "CaseAlt.OptionalSome", type: "Optional[Type]" = None
    ):
        return self.visit_expr_optional_some(
            Expr.OptionalSome(type=type, body=Expr(var=optional_some.var_body))
        )

    def visit_expr_casealt_enum(self, enum: "CaseAlt.Enum"):
        return self.visit_expr_enum_con(Expr.EnumCon(enum.con, enum.constructor))

    def visit_expr_let(self, let: "Block") -> "str":
        pass

    def visit_expr_nil(self, nil: "Expr.Nil") -> "str":
        if self.context.decl_name is not None:
            child = self.with_statement_block()
            return child.visit_expr_nil_decl(nil, self.context)
        else:
            return self.visit_expr_nil_inline(nil)

    def visit_expr_nil_decl(self, nil: "Expr.Nil", context: "CodeContext") -> str:
        expr_text = self.visit_expr_nil_inline(nil)
        return self._visit_expr_decl(context, expr_text)

    def visit_expr_nil_inline(self, nil: "Expr.Nil"):
        raise NotImplementedError

    def visit_expr_cons(self, cons: "Expr.Cons") -> "str":
        pass

    def visit_expr_update(self, update: "Update") -> "str":
        pass

    def visit_expr_scenario(self, scenario: "Scenario") -> "str":
        pass

    def visit_expr_rec_upd(self, rec_upd: "Expr.RecUpd") -> "str":
        pass

    def visit_expr_struct_upd(self, struct_upd: "Expr.StructUpd") -> "str":
        pass

    def visit_expr_optional_none(self, optional_none: "Expr.OptionalNone") -> "str":
        pass

    def visit_expr_optional_some(self, optional_some: "Expr.OptionalSome") -> "str":
        pass

    # noinspection PyBroadException
    def visit_type(self, type: "Type") -> "str":
        """
        Return the string representation of the :class:`Type`, or ``var`` if it could not be
        computed. This implementation tries very hard to not fail.
        """
        default_type = self._visit_type_default()
        try:
            unpacked_types = []

            for unpacked_type in unpack_arrow_type(type):
                try:
                    type_str = super(PrettyPrintBase, self).visit_type(unpacked_type)
                except:  # noqa
                    LOG.exception("Failed to render a Type!")
                    type_str = None
                unpacked_types.append(type_str if type_str is not None else default_type)
            if len(unpacked_types) == 1:
                ret = unpacked_types[0]
            else:
                ret = self._visit_type_function(unpacked_types[0:-1], unpacked_types[-1])
            return ret if ret is not None else default_type
        except:  # noqa
            LOG.exception("Failed to render a Type!")
            return default_type

    def visit_type_var(self, var: "Union[str, Type.Var]") -> "str":
        """
        Render a type variable. The default implementation simply returns the original type
        variable.

        :param var: The :class:`Type.Var` to unpack, or a simple string.
        :return: The rendering of this type variable.
        """
        if isinstance(var, Type.Var):
            return var.var
        elif isinstance(var, str):
            return var
        else:
            raise TypeError("visit_type_var(...) expects either a Type.Var or a str")

    def visit_type_con(self, con: "Type.Con") -> "str":
        pass

    def visit_type_prim(self, prim: "Type.Prim") -> "str":
        pass

    def visit_type_syn(self, tysyn: "Type.Syn") -> str:
        pass

    def visit_type_forall(self, forall: "Type.Forall") -> "str":
        pass

    def visit_type_tuple(self, tuple: "Type.Tuple") -> "str":
        pass

    def visit_type_nat(self, nat: int) -> "str":
        return str(nat)

    def _visit_type_function(self, arguments: "Sequence[Type]", return_type: "Type") -> str:
        """
        Render a function type with the specified arguments and return type.

        :param arguments:
            Arguments to the function type.
        :param return_type:
            The return type of the function.
        :return:
            A string that represents successive type abstractions that, themselves, either return
            further abstractions or a fully applied value.
        """

    def _visit_expr_default(self) -> str:
        """
        Overridden to return a string to use when failing to render an expression.
        """
        return "..."

    def _visit_expr_decl(self, context, expr_text):
        """
        Visit the declaration of a field whose value is specified by a textual string. The default
        implementation simply returns a string of the format ``<name> = <value>``, which happens to
        be the correct implementation in almost all languages.

        :param context:
            The _enclosing_ (not the current) context, which includes information about the name and
            type of the declaration.
        :param expr_text:
            The text to use as the value of the expression.
        :return:
            A string that represents the definition of a value with the name specified by context.
        """
        local_name = ".".join(context.decl_name)
        return f"{local_name} = {expr_text}"

    def _visit_type_default(self) -> str:
        """
        Overridden to return a string to use when failing to render a type.
        """
        return "..."


class CodeContext:
    """
    The context for evaluating a pretty-print expression.
    """

    @classmethod
    def top_level(cls) -> "CodeContext":
        return cls()

    @classmethod
    def from_options(cls, context):
        return cls(show_hidden_types=context.show_hidden_types)

    def __init__(
        self,
        *,
        in_expression: bool = False,
        show_hidden_types: bool = False,
        current_module: Optional[ModuleRef] = None,
        decl_name: Optional[Sequence[str]] = None,
        decl_type: Optional[Type] = None,
        type_abs: Optional[Sequence[TypeVarWithKind]] = None,
        type_app: Optional[Sequence[Type]] = None,
    ):
        self.in_expression = in_expression
        self.show_hidden_types = show_hidden_types
        self.current_module = current_module
        self.decl_name = decl_name
        self.decl_type = decl_type
        self.type_abs = type_abs
        self.type_app = type_app

    def with_module(self, module: ModuleRef) -> "CodeContext":
        if (
            self.decl_name is not None
            or self.decl_type is not None
            or self.type_abs is not None
            or self.type_app is not None
        ):
            raise RuntimeError("Cannot declare a module inside a declaration")
        return CodeContext(
            in_expression=False,
            show_hidden_types=self.show_hidden_types,
            current_module=module,
            decl_name=None,
            decl_type=None,
            type_abs=None,
            type_app=None,
        )

    def with_decl(self, decl_name: Sequence[str], decl_type: "Type") -> "CodeContext":
        return CodeContext(
            in_expression=False,
            show_hidden_types=self.show_hidden_types,
            current_module=self.current_module,
            decl_name=decl_name,
            decl_type=decl_type,
            type_abs=self.type_abs,
            type_app=self.type_app,
        )

    def with_type_abs(self, type_abs: Optional[Sequence[TypeVarWithKind]]):
        return CodeContext(
            in_expression=self.in_expression,
            show_hidden_types=self.show_hidden_types,
            current_module=self.current_module,
            decl_name=self.decl_name,
            decl_type=self.decl_type,
            type_abs=(self.type_abs or ()) + tuple(type_abs),
            type_app=self.type_app,
        )
        pass

    def with_type_app(self, type_app: "Optional[Sequence[Type]]"):
        return CodeContext(
            in_expression=True,
            show_hidden_types=self.show_hidden_types,
            current_module=self.current_module,
            decl_name=None,
            decl_type=None,
            type_abs=None,
            type_app=(self.type_app or ()) + tuple(type_app),
        )

    def with_expression(self):
        return CodeContext(
            in_expression=True,
            show_hidden_types=self.show_hidden_types,
            current_module=self.current_module,
            decl_name=None,
            decl_type=None,
            type_abs=None,
            type_app=None,
        )

    def with_statement_block(self):
        return CodeContext(
            in_expression=False,
            show_hidden_types=self.show_hidden_types,
            current_module=self.current_module,
            decl_name=None,
            decl_type=None,
            type_abs=None,
            type_app=self.type_app,
        )


def decode_special_chars(pp: str) -> str:
    return (
        pp.replace("$u002b", "+")
        .replace("$u005b", "[")
        .replace("$u005d", "]")
        .replace("$u003c", "<")
        .replace("$u003e", ">")
        .replace("$u003a", ":")
        .replace("$u0022", '"')
        .replace("$u0028", "(")
        .replace("$u0029", ")")
        .replace("$u002f", "/")
        .replace("$u002c", ",")
        .replace("$u003d", "=")
        .replace("$u002a", "*")
    )


class _PrettyPrinters:
    """
    Holder for
    """

    def __init__(self):
        self._printers = {}  # type: Dict[str, TType[PrettyPrintBase]]

    # noinspection PyShadowingBuiltins
    def register(self, format: str, pp_type: "TType[PrettyPrintBase]") -> None:
        self._printers[format] = pp_type

    # noinspection PyShadowingBuiltins
    def get(self, format: str) -> "TType[PrettyPrintBase]":
        for key, format_type in self._printers.items():
            if key.startswith(format):
                return format_type


_PRETTY_PRINTERS = _PrettyPrinters()

get_pretty_printer_type = _PRETTY_PRINTERS.get
register_pretty_printer = _PRETTY_PRINTERS.register


def pretty_print_syntax(fmt: str) -> "Callable[[TType[PrettyPrintBase]], TType[PrettyPrintBase]]":
    def impl(pp):
        register_pretty_printer(fmt, pp)
        return pp

    return impl



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from typing import Any, Collection, Mapping, Optional, Set, Union
import warnings

from toposort import toposort_flatten

from ... import LOG
from ...damlast.daml_lf_1 import (
    Archive,
    DefDataType,
    DottedName,
    ModuleRef,
    PackageRef,
    TypeConName,
    ValName,
)
from ...damlast.types import get_old_type

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from ...model.types import (
        SCALAR_TYPE_UNIT,
        EnumType,
        NamedArgumentList,
        RecordType,
        ScalarType,
        Template,
        TemplateChoice,
        TypeReference,
        TypeVariable,
        VariantType,
    )
    from ...model.types_store import PackageStore, PackageStoreBuilder

warnings.warn(
    "The symbols in dazl.protocols.v1.pb_parse_metadata are deprecated",
    DeprecationWarning,
    stacklevel=2,
)


def parse_archive_payload(raw_bytes: bytes, package_id: "Optional[PackageRef]" = None):
    """
    Convert ``bytes`` into a :class:`G.ArchivePayload`.

    Note that this function will temporarily increase Python's recursion limit to handle cases where
    parsing a DAML-LF archive requires deeper recursion limits.
    """
    warnings.warn(
        "Use dazl.damlast.parse.parse_archive_payload instead.", DeprecationWarning, stacklevel=2
    )

    from ...damlast.parse import parse_archive_payload

    return parse_archive_payload(package_id, raw_bytes)


@dataclass(frozen=True)
class ArchiveDependencyResult:
    sorted_archives: "Mapping[PackageRef, G.ArchivePayload]"
    unresolvable_archives: "Mapping[PackageRef, G.ArchivePayload]"


def find_dependencies(
    metadatas_pb: "Mapping[str, G.ArchivePayload]", existing_package_ids: "Collection[PackageRef]"
) -> "ArchiveDependencyResult":
    """
    Return a topologically-sorted list of dependencies for the package IDs.

    :param metadatas_pb:
        The collection of gRPC ArchivePayload to re-order, indexed by package ID.
    :param existing_package_ids:
        Collection of package IDs that are to be treated as pre-existing.
    :return:
        A topologically-sorted dictionary of package IDs to ArchivePayload objects. Iterations
        through the dictionary are guaranteed to be in the correct order.
    """
    warnings.warn(
        "find_dependencies is deprecated; there is no replacement.",
        DeprecationWarning,
        stacklevel=2,
    )

    dependencies = defaultdict(set)
    for package_id, archive_payload in metadatas_pb.items():
        for module_pb in archive_payload.daml_lf_1.modules:
            for data_type_pb in module_pb.data_types:
                deps = None
                if data_type_pb.HasField("record"):
                    deps = find_dependencies_of_fwts(data_type_pb.record.fields)
                elif data_type_pb.HasField("variant"):
                    deps = find_dependencies_of_fwts(data_type_pb.variant.fields)
                if deps is not None:
                    deps.difference_update(existing_package_ids)
                    dependencies[package_id].update(deps)

    # identify any completely missing dependencies
    # TODO: This doesn't handle transitive missing dependencies; this will be a problem once
    #  DAML has proper dependency support
    unresolvable_package_ids = []
    for package_id, package_dependencies in dependencies.items():
        if not package_dependencies.issubset(metadatas_pb):
            unresolvable_package_ids.append(package_id)

    m_pb = {}
    sorted_package_ids = toposort_flatten(dependencies)

    # packages with no dependencies or dependents can safely be added in the front
    remainders = set(metadatas_pb) - set(unresolvable_package_ids) - set(sorted_package_ids)
    sorted_package_ids[0:0] = sorted(remainders)

    for package_id in sorted_package_ids:
        if package_id not in unresolvable_package_ids:
            required_package = metadatas_pb.get(package_id)
            if required_package is not None:
                m_pb[package_id] = required_package
            else:
                LOG.warning("Failed to find a package %r", package_id)

    if unresolvable_package_ids:
        # This isn't a major problem in the grand scheme of things because the caller continually
        # retries all unknown packages, so if the missing dependent packages are eventually
        # uploaded, we will end up back in this function and all of the packages will be parsed.
        LOG.warning(
            "Some package IDs could not be resolved, so packages that depend on these IDs "
            "will be unavailable: %r",
            unresolvable_package_ids,
        )

    return ArchiveDependencyResult(
        sorted_archives=m_pb,
        unresolvable_archives={pkg_id: metadatas_pb[pkg_id] for pkg_id in unresolvable_package_ids},
    )


def find_dependencies_of_fwts(fwts_pb) -> "Set[str]":
    warnings.warn(
        "find_dependencies_of_fwts is deprecated; there is no replacement.",
        DeprecationWarning,
        stacklevel=2,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        dependencies = set()
        for fwt in fwts_pb:
            dependencies.update(find_dependencies_of_fwt(fwt))
        return dependencies


def find_dependencies_of_fwt(fwt_pb) -> "Collection[str]":
    warnings.warn(
        "find_dependencies_of_fwt is deprecated; there is no replacement.",
        DeprecationWarning,
        stacklevel=2,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        return find_dependencies_of_type(fwt_pb.type)


def find_dependencies_of_type(type_pb) -> "Collection[str]":
    warnings.warn(
        "find_dependencies_of_type is deprecated; there is no replacement.",
        DeprecationWarning,
        stacklevel=2,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        t = type_pb.WhichOneof("Sum")  # type: str
        if t == "prim":
            dependencies = set()  # type: Set[str]
            for arg in type_pb.prim.args:
                dependencies.update(find_dependencies_of_type(arg))
            return sorted(dependencies)
        elif t == "con":
            dependencies = set()  # type: Set[str]
            if type_pb.con.tycon.module.package_ref.WhichOneof("Sum") == "package_id":
                dependencies.add(type_pb.con.tycon.module.package_ref.package_id)
            for arg in type_pb.con.args:
                dependencies.update(find_dependencies_of_type(arg))
            return sorted(dependencies)
        elif t == "var":
            dependencies = set()  # type: Set[str]
            for arg in type_pb.var.args:
                dependencies.update(find_dependencies_of_type(arg))
            return sorted(dependencies)
        elif t == "fun":
            dependencies = set()  # type: Set[str]
            for arg in type_pb.fun.params:
                dependencies.update(find_dependencies_of_type(arg))
            dependencies.update(find_dependencies_of_type(type_pb.fun.result))
            return dependencies
        elif t == "forall":
            return find_dependencies_of_type(type_pb.forall.body)
        elif t == "tuple":
            return find_dependencies_of_fwts(type_pb.tuple.fields)
        elif t == "nat":
            return ()
        else:
            LOG.warning("Unknown DAML-LF Type: %s (when evaluating %s)", t, type_pb)
            return ()


def parse_daml_metadata_pb(package_id: "PackageRef", metadata_pb: Any) -> "PackageStore":
    """
    Parse the contents of the given DAML-LF archive.

    :param package_id:
        The package to associate "self"-modules to.
    :param metadata_pb:
        The Protobuf-generated Archive object.
    :return:
        A :class:`PackageStore` with additional entries resulting from the parse of this archive.
    """
    warnings.warn(
        "parse_daml_metadata_pb and PackageStore are deprecated; "
        "use dazl.damlast.parse_archive instead",
        DeprecationWarning,
        stacklevel=2,
    )

    LOG.debug("Parsing package ID: %r", package_id)

    from ...damlast.pb_parse import ProtobufParser

    parser = ProtobufParser(package_id)
    package = parser.parse_Package(metadata_pb.daml_lf_1)

    return _parse_daml_metadata_pb(Archive(package_id, package))


def _parse_daml_metadata_pb(archive: "Archive") -> "PackageStore":
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        psb = PackageStoreBuilder()
        psb.add_archive(archive)

        for module in archive.package.modules:
            current_module = ModuleRef(archive.hash, DottedName(module.name.segments))
            for vv in module.values:
                vt = ValName(current_module, vv.name_with_type.name)
                psb.add_value(vt, vv.expr)

            for dt in module.data_types:
                tt = create_data_type(current_module, dt)
                if isinstance(tt, (RecordType, VariantType, EnumType)):
                    psb.add_type(tt.name.con, tt)
                else:
                    LOG.warning("Unexpected non-complex type will be ignored: %r", tt)

            for template_pb in module.templates:
                con = TypeConName(current_module, template_pb.tycon.segments)
                data_type = psb.get_type(con)
                if isinstance(data_type, RecordType):
                    psb.add_template(
                        Template(
                            data_type=data_type,
                            key_type=get_old_type(template_pb.key.type)
                            if template_pb.key is not None
                            else None,
                            choices=[
                                TemplateChoice(
                                    c.name,
                                    c.consuming,
                                    get_old_type(c.arg_binder.type),
                                    get_old_type(c.ret_type),
                                    c.controllers,
                                )
                                for c in template_pb.choices
                            ],
                            observers=template_pb.observers,
                            signatories=template_pb.signatories,
                            agreement=template_pb.agreement,
                            ensure=template_pb.precond,
                        )
                    )
                elif data_type is None:
                    LOG.warning(
                        "The template %s did not have a corresponding data definition; "
                        "it will be ignored",
                        con,
                    )
                else:
                    LOG.warning(
                        "The template %s was of type %s; only records are supported for templates",
                        con,
                        data_type,
                    )

        LOG.debug("Fully registered all types for package ID %r", archive.hash)
        return psb.build()


def create_data_type(
    current_module_ref: "ModuleRef", dt: "DefDataType"
) -> "Union[RecordType, VariantType, EnumType, ScalarType]":
    warnings.warn(
        "create_data_type is deprecated; there is no replacement.", DeprecationWarning, stacklevel=2
    )

    from ...damlast.types import get_old_type

    type_vars = tuple(TypeVariable(type_var.var) for type_var in dt.params)
    tt = TypeReference(con=TypeConName(current_module_ref, dt.name.segments))

    if dt.record is not None:
        d = OrderedDict()
        for fwt in dt.record.fields:
            d[fwt.field] = get_old_type(fwt.type)
        return RecordType(NamedArgumentList(d.items()), tt, type_vars)
    elif dt.variant is not None:
        d = OrderedDict()
        for fwt in dt.variant.fields:
            d[fwt.field] = get_old_type(fwt.type)
        return VariantType(NamedArgumentList(d.items()), tt, type_vars)
    elif dt.enum is not None:
        return EnumType(tt, dt.enum.constructors)
    else:
        return SCALAR_TYPE_UNIT
